# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
import matplotlib.pyplot as plt
# %matplotlib inline
from matplotlib.colors import LogNorm
from mpl_toolkits.mplot3d import axes3d
import matplotlib.pyplot as plt

# %%
import os
import numpy
from astropy.io import fits
import shutil
import tempfile

from micado.config.Environment import Env

from micado.util.esokeywords import verify_fitsheaders

from micado.main.Img import Img
from micado.main.ObservingBlock import ObservingBlock
from micado.main.Imstat import Imstat

from micado.hardware.ShotNoise import ShotNoise
from micado.hardware.SummedExposure import SummedExposure
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.hardware.LinearityCurve import LinearityCurve
from micado.hardware.QuantumEfficiencyCurve import QuantumEfficiencyCurve
from micado.hardware.SurfaceList import SurfaceList
from micado.hardware.Instrument import Instrument
from micado.hardware.RelayOptics import RelayOptics
from micado.hardware.FieldConstantPSF import FieldConstantPSF
from micado.hardware.FilterWheel import FilterWheel
from micado.hardware.SummedExposure import SummedExposure
from micado.hardware.NonCommonPathAberration import NonCommonPathAberration
from micado.hardware.Atmosphere import Atmosphere
from micado.hardware.AtmosphericTERCurve import AtmosphericTERCurve
from micado.hardware.AtmosphericDispersion import AtmosphericDispersion
from micado.hardware.Vibration import Vibration
from micado.hardware.Telescope import Telescope

from micado.dataitems.Dark import Dark
from micado.dataitems.Dark import DarkExtension
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterDark import MasterDarkExtension
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.MasterFlatExtension import MasterFlatExtension
from micado.dataitems.Flat import Flat
from micado.dataitems.FlatExtension import FlatExtension

from micado.dataitems.Science import Science
from micado.dataitems.ScienceExtension import ScienceExtension

from micado.hardware.LinearityCurve import default_linearitycurve
from micado.hardware.QuantumEfficiencyCurve import default_quantumefficiencycurve
from micado.hardware.ReferencePixelBorder import default_referencepixelborder

from micado.requirements.DPSmasterdark import DPSmasterdark

from micado.templates.MicadoImgCalFlat import MicadoImgCalFlat
from micado.templates.MicadoImgCalDark import MicadoImgCalDark
from micado.templates.MicadoImgObs import MicadoImgObs

from micado.recipes.McdImgFlat import McdImgFlat
from micado.recipes.McdDetDark import McdDetDark

from micado.test.functional.dark_test import remove_file_and_links, compare_raws
from micado.test.functional.flat_test import generate_flat
from micado.test.functional.masterdark_test import (
    compare_pros,
    create_masterdark,
)

from synphot import SourceSpectrum, BlackBody1D
from scopesim.effects.ter_curves_utils import scale_spectrum
from scopesim import Source
import scopesim_templates


# %%
def generate_flat(dark: float = 2.0, hw: int = 32):
    """Generate a Flat.

    Args:
        dark: Value of the dark_current.
        hw: Size of the detector.
    
    Based on micado.test.functional.flat_test.generate_flat()
    """
    filt1 = FilterWheel(
        name="Ks",
        minimum_throughput=0.000101,
        outer=0.2,
    )
    filt2 = FilterWheel(
        name="open",
        minimum_throughput=0.000101,
        outer=0.2,
    )
    inst = Instrument(
        name="MICADO",
        ocs__pxscale=0.004,
        plate_scale=0.26666666666,
        filter_wheel_1=filt1,
        filter_wheel_2=filt2,
        static_surfaces=SurfaceList(realfilename="LIST_MICADO_mirrors_static.dat"),
        ncpas_psf=NonCommonPathAberration(realfilename='INST_MICADO_wavefront_error_budget.dat'),
    )
    ro = RelayOptics(
        psf=FieldConstantPSF(
            realfilename="PSF_SCAO_ConstPSF_0_5off.fits",
        ),
        surface_list=SurfaceList(
            realfilename="LIST_RO_SCAO_mirrors.dat",
        ),
    )    
    
    mydetector = Detector(x_size=hw, y_size=hw,)
    mydetectorlist = [mydetector]
    mydarkcurrent = DarkCurrent(value=dark)
    # mydarkcurrent = None
    myshotnoise = ShotNoise(state=1)
    mydetectorarray = DetectorArray(
        name="",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        qe_curve=default_quantumefficiencycurve,
        linearity=default_linearitycurve,
        reference_pixel_border=default_referencepixelborder,
        exposure_action=SummedExposure(state=1),
        default_objects=False,
    )

    flat_template = MicadoImgCalFlat(det__dit=1, det__ndit=1)

    flat_ob = ObservingBlock(
        name="MyFlatTest",
        instrument="MICADO",
        target="flatlamp",
        templates=[flat_template],
    )

    myextensions = [
        FlatExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
    ]

    flat = Flat(
        detector=mydetectorarray,
        template=flat_template,
        observing_block=flat_ob,
        extensions=myextensions,
        instrument=inst,
        relay_optics=ro,
    )
    flat.make_simulation()
    return flat



# %%
# Create a MasterFlat from 1 Flat of 16x16 pixels when simple is set.
n_flats = 5
# hw = 0.24
hw = 4096
dark = .1

# %%
myflat = generate_flat(dark=dark, hw=hw)

# %%
data = myflat.extensions[0].image
dmin, dmax, dmean, dmed, dstd = data.min(), data.max(), data.mean(), numpy.median(data), data.std()
fig, ax = plt.subplots()
im = ax.imshow(data, norm=LogNorm(vmin=dmed , vmax=dmed + 0.1 * dstd))
# im = ax.imshow(data)
fig.colorbar(im)
fig.show()


# %%
print(dmin, dmax, dmean, dmed, dstd)
print(dmed , dmed + 0.1 * dstd)
print(data.shape)

# %%

# %% [markdown]
# ## MasterFlat

# %%
mydetector = Detector(x_size=hw, y_size=hw,)
mydetectorlist = [mydetector]

# Create raw Darks.
myflats = []
for _ in range(n_flats):
    myflat = generate_flat(dark=dark, hw=hw)
    myflat.make_simulation()
    myflat.save_fits()

    myflats.append(myflat)

mymasterdark, _, _ = create_masterdark()

myrecipe = McdImgFlat()

mymasterextensions = [
    MasterFlatExtension(BITPIX=-64, NAXIS=2) for _ in mydetectorlist
]

mymasterflat = MasterFlat(
    recipe=myrecipe,
    flats=myflats,
    masterdark=mymasterdark,
    extensions=mymasterextensions,
)
mymasterflat.make()
mymasterflat.save_fits()


# %%
myimstatraw = Imstat(baseframe=myflats[0])
myimstatraw.make()

myimstat = Imstat(baseframe=mymasterflat)
myimstat.make()

# The standard deviation cannot be used to measure the
# effectiveness of the MasterFlat routine, because the
# structure of the flat means that the noise does not 
# scale with sqrt(n_raws) over the full image.
# assert myimstat.std < myimstatraw.std / 1.5

print(myimstat.std, myimstatraw.std, myimstatraw.std / 1.5, myimstat.std < myimstatraw.std / 1.5)


# %%
data = mymasterflat.extensions[0].image
dmin, dmax, dmean, dmed, dstd = data.min(), data.max(), data.mean(), numpy.median(data), data.std()
fig, ax = plt.subplots()
im = ax.imshow(data)
# im = ax.imshow(data, norm=LogNorm(vmin=dmed , vmax=dmed + 0.1 * dstd))
fig.colorbar(im)
fig.show()



# %%
print(data)
print(dmin, dmax, dmean, dmed, dstd)

# %%

# %%
# There is an off-by-one bug in ScopeSIM, so need to crop the image.
data_plot = data[1:-1, 1:-1]

# %%
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')

# Grab some test data.
X = numpy.arange(data_plot.shape[0])
Y = numpy.arange(data_plot.shape[1])
X, Y = numpy.meshgrid(X, Y)
Z = data_plot

# Plot a basic wireframe.
ax.plot_wireframe(X, Y, Z, rstride=10, cstride=10)

plt.show()

# %%
