# -*- coding: utf-8 -*-
"""The BaseFrame is the base class for all image-based classes.

It represents a single-layer image.
"""

import logging
import tempfile

from astropy.io import fits

from micado.main.DataItem import DataItem


class BaseFrameExtra(DataItem):
    """BaseFrame."""

    # TODO: use these time formats again

    format_time_filename = "%Y%m%dT%H%M%SS%f"
    # Format for strftime/strptime for filenames

    format_time_fitsheader = "%Y-%m-%dT%H:%M:%S.%f"
    # Format for strftime/strptime for fits headers

    header_org: fits.Header = None
    # The original FITS header, if any.

    @classmethod
    def compare_fits_header_entry(cls, name, value, value_compare):
        """compare the fits header entry with name and value against value_compare"""
        if name == "FNSRC0" and "(" in value_compare:
            # the called methods on scopesim can contain arguments that are not set
            # on the WISE side, remove the arguments from the ()
            element1, element2 = value_compare.rsplit("(", 1)
            value_compare = element1 + "()"
        return value == value_compare

    def save_fits(self, filename: str = None):
        """Save as FITS file."""
        # TODO: Ensure the path where FITS files are stored and read from
        # can be set.
        if not filename:
            if not self.filename:
                self.filename = tempfile.NamedTemporaryFile(
                    dir=".",
                    prefix="MCDWS_",
                    suffix=".fits",
                    delete=False,
                ).name
            filename = self.filename
        logging.getLogger(__name__).info(f"save_fits to {filename}")
        header_primary_auto = fits.Header(self.fits_keywords)
        if self.header_org:
            header_primary = self.header_org.copy()
            # Verify that the ScopeSIM header items are what MicadoWISE
            # expects to be there and add header items that are missing.
            # E.g. ScopeSIM adds HIERARCH ESO DPR CATG, which means that it
            # is not necessary to add HIERARCH WISE DPR CATG, but the values
            # should be the same.
            for key, value in header_primary_auto.items():
                key2 = key.replace("WISE ", "ESO ")
                if key in header_primary:
                    assert self.compare_fits_header_entry(
                        key, value, header_primary[key]
                    ), f"{key}: {value} != {header_primary[key]}"
                elif key2 in header_primary:
                    assert self.compare_fits_header_entry(
                        key, value, header_primary[key2]
                    ), f"{key2}: {value} != {header_primary[key2]}"
                else:
                    # TODO: raise an exception here once ScopeSIM can produce
                    # the exact headers that are required.
                    header_primary[key] = value
            # header_primary.update(header_primary_auto)
        else:
            header_primary = header_primary_auto
        hduprimary = fits.PrimaryHDU(header=header_primary)

        # TODO: Use the same header verification mechanism as for the primary.
        hdusextension = []
        for exti, ext in enumerate(self.extensions, 1):
            header_ext_auto = fits.Header(ext.fits_keywords)
            if ext.header_org:
                header = ext.header_org.copy()
                header.update(header_ext_auto)
            else:
                header = header_ext_auto

            from micado.main.Raw import Raw

            if ext.image is not None:
                # TODO: This should also happen for Raw data! Fix with ScopeSIM.
                if not isinstance(self, Raw):
                    # header = fits.Header(ext.fits_keywords)
                    header["EXTNAME"] = f"DET{exti}.DATA"
                    header["HDUCLAS2"] = "DATA"
                    header["HDUCLAS3"] = ""
                hdu = fits.ImageHDU(
                    data=ext.image,
                    header=header,
                )
                hdusextension.append(hdu)
            if ext.uncertainty is not None:
                header = fits.Header()
                header["EXTNAME"] = f"DET{exti}.ERR"
                header["HDUCLAS2"] = "ERROR"
                header["HDUCLAS3"] = "RMSE"
                hdu = fits.ImageHDU(
                    data=ext.uncertainty,
                    header=header,
                )
                hdusextension.append(hdu)
            if ext.quality is not None:
                header = fits.Header()
                header["EXTNAME"] = f"DET{exti}.DQ"
                header["HDUCLAS2"] = "QUALITY"
                header["HDUCLAS3"] = "FLAG32BIT"
                hdu = fits.ImageHDU(
                    data=ext.quality,
                    header=header,
                )
                hdusextension.append(hdu)

        hdul = fits.HDUList([hduprimary] + hdusextension)
        hdul.writeto(filename, output_verify="fix", overwrite=True)

        self.unique_filename()

    def unique_filename(self, input_filename=None, extra=None):
        """Generate unique filename."""
        extra = extra if extra is not None else []
        # TODO: The extra is currently ignored in order to put the filenames
        # in the FITS headers. The extra's should be reenabled or removed
        # entirely.
        DataItem.unique_filename(
            self,
            input_filename=input_filename,
            extra=[self.creation_date.strftime(self.format_time_filename)] + extra,
        )
