# -*- coding: utf-8 -*-
"""Add some properties to DBObject."""
import os
import re
from collections import OrderedDict
from enum import Enum
from typing import List, Tuple, Union

from astropy.io import fits

from micado.main.MDBObjectMeta import MDBObjectMeta

from micado.main.commonwrapper import DBObject
from micado.main.commonwrapper import link_property, persistent_property
from micado.main.commonwrapper import DataObject


def fix_key(key):
    """Make HIERARCH keywords and fix Template parameters."""
    # A # is used as placeholder for i, which is a placeholder for a number,
    # usually the detector number.
    key = key.replace("#", "i")
    if "__" in key:
        # Workaround for template parameters
        elements = [kk for kk in key.split() if "__" in kk]
        assert len(elements) == 1
        element = elements[0]
        key = element.replace("__", " ")
    if " " in key:
        if "PRO" in key:
            # Always use ESO compliance for PRO keywords.
            key = f"HIERARCH ESO {key}"
        # else:
        elif not key.startswith("SIM"):
            # Some of the ScopeSIM keywords can be reused as is.
            key = f"HIERARCH WISE {key}"
    # TODO: Generalize this translation to other attributes.
    if key == "MJDOBS":
        key = "MJD-OBS"
    return key


def classes_recurse_fits_headers():
    """Classes to recurse into when generating FITS headers.

    Cannot be imported at the top-level because of circular imports.
    """
    # TODO: Add Recipe back here and overload get_fits_keywords(_c) with
    #       the get_parameters stuff.
    # TODO: Perhaps optionally add DataItem back? It is removed because it
    #       should never be necessary. DataItem should only occur for
    #       data lineage dependencies of processed data, and should thus
    #       be encoded in an ESO-compliant way.
    from micado.main.Hardware import Hardware
    from micado.main.Template import Template
    from micado.main.ObservingBlock import ObservingBlock
    from micado.dataitems.QCParameters import QCParameters
    from micado.main.Recipe import Recipe

    return Hardware, Template, ObservingBlock, QCParameters, Recipe


def getkw(attrc, prefix=None):
    """Recursively get keyword attributes."""
    from micado.main.DataItem import DataItem

    classes_ok = classes_recurse_fits_headers()
    prefix = [] if prefix is None else prefix
    if isinstance(attrc, DataItem):
        # This is a dependency.
        if attrc.filename:
            prefix2a = prefix + ["filename"]
            ww = OrderedDict([(" ".join(prefix2a).upper(), attrc.filename)])
        else:
            ww = OrderedDict([(" ".join(prefix).upper(), "")])
    elif isinstance(attrc, classes_ok):
        ww = attrc.real_fits_keywords(prefixes=prefix)
    elif isinstance(attrc, Enum):
        # TODO: Add these enums to the templates
        # DprCategory or DprType
        # These are converted to strings for the time being.
        # TODO: Ensure database supports Enums and Flags
        # E.g. 'DprCategory.CALIB|SCIENCE' -> 'CALIB|SCIENCE'
        val = str(attrc).split(".")[-1]
        ww = OrderedDict([(" ".join(prefix).upper(), val)])
    else:
        ww = OrderedDict([(" ".join(prefix).upper(), attrc)])
    return ww


def getkw_c(attrc, prefix=None):
    """Recursively get keyword attributes.

    Same as getkw, but on the class level and only returns the keywords.
    """
    from micado.main.DataItem import DataItem

    classes_ok = classes_recurse_fits_headers()
    prefix = [] if prefix is None else prefix
    assert "PRO" not in prefix
    if issubclass(attrc, DataItem):
        prefix2a = prefix + ["filename"]
        return [" ".join(prefix2a).upper()]
    elif issubclass(attrc, classes_ok):
        return attrc.real_fits_keywords_c(prefixes=prefix)
    else:
        return [" ".join(prefix).upper()]


def shortname_from_attrname(attrname):
    """Convert long attribute name to short name for FITS keywords."""
    # Search for :3 to find more of such cases. TODO: Harmonize.
    attrnameclean = attrname.replace("_", "")
    attrname_short = "{:.3}".format(attrnameclean)
    # TODO: Make more general solution for this ugly TEM/TPL hack.
    if attrname_short == "tem":
        attrname_short = "tpl"
    # TODO: Make more general solution for this ugly FILT hack.
    if attrname_short == "fil":
        attrname_short = "filt"
    # TODO: Make more general solution for this ugly PUP / OPTI6 hack.
    # TODO: This produces the incorrect 'HIERARCH ESO OPTIi INS FILT NAME'
    #       Which is hacked around in combine_prefixes_with_esokeyword()
    if attrname_short == "pup":
        attrname_short = "opti6"
    # E.g. there is INS.FILT1 and INS.FILT2.
    if attrname[-1] in "123456789":
        attrname_short += attrnameclean[-1]
    return attrname_short


def combine_prefixes_with_esokeyword(prefixes, esokeyword):
    """Combine prefixes with esokeyword."""
    # If prefixes is ["ins", "mod", "sur#"] and
    # esokeyword is "SUR INNER", then
    # add "INS MOD SURi INNER". That is, remove the redundant SUR.
    # Also prevent "INS FILT2 INS FILT NAME"
    # TODO: Fix this hack to prevent "INS OPTIi INS FILT NAME"
    if "opti6" in prefixes:
        esokeyword = esokeyword.replace("FILT", "OPTI6")
    if prefixes:
        prefix_last = prefixes[-1]
        prefix_last_s = prefix_last[:-1] + " "
        # TODO: This prefix_last_s.upper() in esokeyword test is
        # error prone.
        if prefix_last[-1] in "#1234567890" and prefix_last_s.upper() in esokeyword:
            pos_new = esokeyword.find(prefix_last_s.upper()) + len(prefix_last)
            esokeyword = esokeyword[pos_new:]
    # If prefixes is ["abc", "det", "det#"] and
    # esokeyword is "DET DETi XYZ" then
    # add "ABC DET DETi XYZ".
    prefix2 = " ".join(prefixes).upper().replace("#", "i")
    # TODO: Fix this hack to prevent "INS OPTIi INS FILT NAME"
    prefix2 = prefix2.replace("OPTI6", "OPTISIX")
    for num in "0123456789":
        prefix2 = prefix2.replace(num, "i")
    prefix2 = prefix2.replace("OPTISIX", "OPTI6")
    for i, _ in enumerate(prefix2):
        prefix3 = prefix2[:i].strip()
        prefix_to_ignore = prefix2[i:]
        if esokeyword.startswith(prefix_to_ignore):
            break
    else:
        prefix3 = prefix2

    # Do not combine DPR and TPL special cases.
    # That is, ["obs", "tpl#"] + "DPR CATG" -> "DPR CATG"
    # And ["tpl"] + "INS MODE" -> "INS MODE".
    # However, allow ["tpl"] + "START" -> "TPL START".
    # TODO: Find generic way to decide when to combine or not.
    is_dpr_in_esokeyword = "DPR" in esokeyword
    is_esokeyword_subsystem = " " in esokeyword or "." in esokeyword
    is_tpl_in_prefix = "TPL" in prefix3
    if (
        not prefix3
        or is_dpr_in_esokeyword
        or is_tpl_in_prefix
        and is_esokeyword_subsystem
    ):
        kw = esokeyword
    else:
        kw = " ".join([prefix3] + [esokeyword])
    return kw


def int_from_float_if_identical(afloat: float) -> Union[float, int]:
    """Convert a float to an int if the value stays identical."""
    return int(afloat) if int(afloat) == afloat else afloat


class MObjectExtra(DataObject, metaclass=MDBObjectMeta):
    """Enhanced DBObject."""

    def commit(self):
        """Overloaded to prevent warning about storing.

        Storing of MDBObjects is not strictly necessary, because they can
        usually be reprocessed from their dependencies.
        """
        if DataObject.storage:
            # WISE-common
            obj = DataObject.storage.object_to_commit(self)
        else:
            # mock-common
            obj = self
        DBObject.commit(obj)

    @classmethod
    def fits_keywords_c(cls):
        """Get FITS keywords and default values for this class."""
        from micado.main.Recipe import Recipe
        from micado.main.RecipeStepExtra import name_eso_from_property

        kws = cls.real_fits_keywords_c()
        if hasattr(cls, "recipe"):
            assert issubclass(cls.recipe.prop_type, Recipe)
            # TODO: Properly support Recipe's.
            # kws += ["REC NAME"]
            kws += [
                name_eso_from_property(prop)
                for prop in cls.recipe.prop_type.get_parameters_c()
            ]

        return [fix_key(k) for k in kws]

    @classmethod
    def real_fits_keywords_c(cls, prefixes=None):
        """The lower level function that is called recursively."""
        from micado.main.RecipeStepExtra import name_eso_from_property
        from micado.main.DataItem import DataItem
        from micado.hardware.Detector import Detector

        # TODO: Move to DetectorExtra
        # Overloaded to match ScopeSIM.
        if cls == Detector:
            return [
                "ID",
                "ANGLE",
                "GAIN",
                "pixel_size",
                "X_CEN",
                "Y_CEN",
                "X_SIZE",
                "Y_SIZE",
                "FILENAME",
            ]

        classes_ok = classes_recurse_fits_headers()

        prefixes = [] if prefixes is None else prefixes
        proptypes1 = cls.get_proptypes_for_fitsheader()
        # DataItems should not be normal header keywords.
        # TODO: Add RECi keywords.
        # Ignore realfilename on the class level.
        # It is necessary on the instance level though.
        # TODO: Fix realfilename hack.
        proptypes = {
            k: v
            for k, v in proptypes1.items()
            if k not in ["realfilename", "recipe"] and not issubclass(v, DataItem)
        }
        attrnames = proptypes.keys()

        attrs = []
        for attrname in attrnames:
            attr = cls.__getattribute__(cls, attrname)
            proptype = proptypes[attrname]
            esokeyword = name_eso_from_property(attr)
            attrname_short = shortname_from_attrname(attrname)
            if isinstance(attr.default, list):
                prefix2 = prefixes + ["{}#".format(attrname_short)]
                attrs += getkw_c(proptype, prefix=prefix2)
            elif issubclass(proptype, classes_ok):
                prefix2 = prefixes + [attrname_short]
                attrs += getkw_c(proptype, prefix=prefix2)
            elif esokeyword and esokeyword != "None":
                # There is an explicit ESO keyword given in the description.
                assert "PRO" not in esokeyword
                kw = combine_prefixes_with_esokeyword(prefixes, esokeyword)
                attrs += [kw]
            else:
                assert attrname in ["filename"]
                # filename is inherited from common DataObject and does not
                # have to be in the FITS header.
            assert all("PRO" not in a for a in attrs)
        # Make keys unique through OrderedDict
        return list(OrderedDict((k, True) for k in attrs).keys())

    @classmethod
    def get_proptypes_for_fitsheader(cls):
        """Attribute: Property type dictionary to construct FITS headers."""
        from micado.main.Simulator import SimulatorRun
        from micado.main.RecipeStepExtra import RecipeStepBase

        proptypes1 = {
            an: getattr(cls, an).prop_type for an in cls.get_persistent_properties()
        }
        proptypes2 = {
            # sometimes prop_type is (float, int)
            an: pt[0] if isinstance(pt, tuple) else pt
            for an, pt in proptypes1.items()
        }
        from micado.dataitems.QCParameters import QCParameters

        return {
            attrnamea: proptype
            for attrnamea, proptype in proptypes2.items()
            # TODO: Support Requirement somehow?
            # TODO: Better support Recipe formally.
            if attrnamea not in ["object_id", "extensions", "globalname", "requirement"]
            and not issubclass(proptype, SimulatorRun)
            and not (
                # The filename of Recipe is allowed because Recipe parameters
                # are not working at all. Having the filename makes the
                # Recipe at least be instantiated.
                issubclass(cls, (RecipeStepBase, QCParameters))
                and attrnamea in ["filename"]
            )
            # and attrnamea not in ["realfilename"]
        }

    @property
    def fits_keywords(self):
        """Create FITS keywords."""
        kws = self.real_fits_keywords()
        return OrderedDict([(fix_key(k), v) for k, v in kws.items()])

    def real_fits_keywords(self, prefixes=None):
        """Real FITS kws."""
        from micado.main.MObject import MObject
        from micado.main.RecipeStepExtra import name_eso_from_property

        classes_ok = classes_recurse_fits_headers()
        # TODO: Fix this PupilWheel hack
        from micado.hardware.PupilWheel import PupilWheel
        from micado.main.DataItem import DataItem
        from micado.hardware.Detector import Detector

        # TODO: Move to DetectorExtra.
        # Overloaded to match ScopeSIM.
        if isinstance(self, Detector):
            # ScopeSIM seems to use integers for several values.
            x_cen = int_from_float_if_identical(self.x_cen)
            y_cen = int_from_float_if_identical(self.y_cen)
            gain = int_from_float_if_identical(self.gain)
            angle = int_from_float_if_identical(self.angle)
            # For some reason ScopeSIM uses strings for everything.
            # TODO: Fix pixel_size hack.
            # TODO: Fix str hack.
            return OrderedDict(
                [
                    ("ID", str(self.id)),
                    ("ANGLE", str(angle)),
                    ("GAIN", str(gain)),
                    ("pixel_size", self.pixel_size),
                    ("X_CEN", str(x_cen)),
                    ("Y_CEN", str(y_cen)),
                    ("X_SIZE", str(self.x_size)),
                    ("Y_SIZE", str(self.y_size)),
                ]
            )

        # Do not recurse into PupilWheel because it is not supported.
        classes_notok = (PupilWheel,)

        prefixes = [] if prefixes is None else prefixes
        proptypes = self.get_proptypes_for_fitsheader()
        attrnames = proptypes.keys()

        attrs = OrderedDict()
        for attrname in attrnames:
            attr: MObject = self.__getattribute__(attrname)
            proptype = proptypes[attrname]
            attrname_short = shortname_from_attrname(attrname)
            if isinstance(attr, list) and issubclass(proptype, classes_ok):
                # Skip list of Detectors, those should be fetched from the
                # extensions instead.
                if not proptype == Detector:
                    for attri, attr2 in enumerate(attr):
                        prefix2 = prefixes + ["{}{}".format(attrname_short, attri + 1)]
                        attrs_new = getkw(attr2, prefix=prefix2)
                        attrs.update(attrs_new)
            elif issubclass(proptype, classes_ok):
                if not issubclass(proptype, classes_notok):
                    if attr is not None:
                        prefix2 = prefixes + [attrname_short]
                        attrs.update(getkw(attr, prefix=prefix2))
            elif not issubclass(proptype, DataItem):
                # Convert 'instrument' to 'INSTRUME' but keep 'det__ndit'.
                prop = getattr(self.__class__, attrname)
                name_eso = name_eso_from_property(prop, auto=False)
                assert name_eso.lower() != "none", name_eso
                assert name_eso.lower() != "state", name_eso
                # TODO: Fix yet another place for MJDOBS <-> MJD-OBS hack
                if attrname == "mjdobs":
                    attrname8 = "mjd-obs"
                    prefix2 = prefixes + [attrname8]
                    attrs.update(getkw(attr, prefix=prefix2))
                elif attrname == "filename":
                    # This is not a DataItem, so probably Hardware. There are
                    # three options that are legitimate:
                    # 1) Other attributes are set, but not the filename.
                    #    E.g. the DetectorArray can be explicitly specified,
                    #    or it can just be FPA_array_layout.dat.
                    #    Well that example probably does not work because
                    #    it is a list.
                    # 2) The filename is set, but none of the attributes.
                    #    So one cannot have a filename *and* attributes.
                    #    That is, the file should contain the attributes.
                    # 3) An empty file, if there are no attributes.
                    #    Is this even necessary?
                    #
                    # TODO: The above doesn't work at all. That is why
                    #       there currently is also realfilename, fix this.
                    pass
                elif "__" in attrname:
                    # TODO: This is ridiculous going back and forth like this.
                    # E.g. ins__filt1__name should stay the way it is,
                    # but spectrum__filename should become ATM SPE FILENAME.
                    # So it here belowe becoms atm__spe__filename so it is
                    # again fixed with fix_key later....
                    # All attributes with __ have an eso name.
                    assert name_eso, attrname
                    attrname8 = name_eso.replace(" ", "__").replace(".", "__").lower()
                    prefix2 = prefixes + [attrname8]
                    attrs.update(getkw(attr, prefix=prefix2))
                else:
                    # TODO: This is a really ugly hack to get DET DET2 ANGLE
                    #       etc. to work.
                    if name_eso and (["X"] + prefixes)[-1][-1] not in "0123456789":
                        attrname8 = name_eso.replace(".", " ")
                        attrs[attrname8] = attr
                    else:
                        prefix_last = (["UGLYHACK"] + prefixes)[-1].strip("0123456789")
                        prefix_2nd_last = (["UGLY", "HACK"] + prefixes)[-2].strip(
                            "0123456789"
                        )
                        if name_eso and name_eso.startswith(prefix_last.upper()):
                            # E.g. prefixes = ["DET", "DET1"], name_eso = "DET PIXELSIZ"
                            attrname8 = name_eso.split(maxsplit=1)[-1]
                        elif name_eso and name_eso.startswith(
                            f"{prefix_2nd_last} {prefix_last}".upper()
                        ):
                            # E.g. prefixes = ["INS", "FILT1"], name_eso = "INS FILT NAME"
                            attrname8 = name_eso.split(maxsplit=2)[-1]
                        elif name_eso:
                            attrname8 = name_eso
                        else:
                            attrname8 = attrname[:8]
                        prefix2 = prefixes + [attrname8]
                        attrs.update(getkw(attr, prefix=prefix2))

        # TODO: Ensure this works for processed catalogs as well.
        from micado.main.Img import Img

        if isinstance(self, Img):
            attrs.update(self.pro_fits_keywords())

        # Remove FILENAME, because it is calculated from the data.
        # See dark_test::test_resave
        if "FILENAME" in attrs:
            attrs.pop("FILENAME")

        return attrs

    def pro_fits_keywords(self):
        """Get header keywords for processed data recursively.

        See ESO-VLT-DIC.PRO for details on how the data lineage is encoded.
        """
        attr_prog, cls_prog = self.__class__.get_progenitor_attr_class()
        progenitors = getattr(self, attr_prog.attribute)
        if not isinstance(progenitors, list):
            progenitors = [progenitors]
        progenitor_main = progenitors[0]
        # TODO: This does not work for processed catalogs because those are
        #       not Img! Add processed base class.
        from micado.main.Img import Img

        pro_headers: List[List[Tuple[str, str]]] = []
        if isinstance(cls_prog, Img):
            # The progenitor is processed as well.
            pro_headers += progenitor_main.pro_fits_headers()

        nr_recipe = len(pro_headers) + 1
        # TODO: Add CALIB
        # TODO: Harmonize with get_sof_dict()
        headers1 = [
            [
                (f"PRO REC{nr_recipe} RAW{i} NAME", prog.filename),
                (f"PRO REC{nr_recipe} RAW{i} CATG", cls_prog.name_oca()),
            ]
            for i, prog in enumerate(progenitors, 1)
        ]
        # flatten
        headers = [h for hs in headers1 for h in hs]
        pro_headers += headers

        attrcalib = self.__class__.get_calibdata()
        headers_calib = []
        for i, (attr_calib, cls_calib) in enumerate(attrcalib.items(), 1):
            calib = getattr(self, attr_calib.attribute)
            # TODO: Should the numbering continue when calib is None or not?
            if calib is not None:
                headers_calib += [
                    (f"PRO REC{nr_recipe} CALIB{i} NAME", calib.filename),
                    (f"PRO REC{nr_recipe} CALIB{i} CATG", cls_calib.name_oca()),
                ]

        pro_headers += headers_calib

        return OrderedDict(pro_headers)

    @property
    def hierarch_shortnames(self):
        """2-way dictionary with short names for persistent attributes.

        E.g. for Dark
        {'detector': 'DET',
         'observing_block': 'OBS',
         'simulator': 'SIM',
         'template': 'TEM',
         'DET': 'detector',
         'OBS': 'observing_block',
         'SIM': 'simulator',
         'TEM': 'template'}

        Removes _, e.g.
        DetectorArray.qe_curve -> DET QEC
        """
        # TODO: use this function somehowe, e.g. TEM should become TPL
        attrnames = self.__class__.get_persistent_properties()
        props = [getattr(self.__class__, an) for an in attrnames]
        attrshierarch = [
            (a, a.replace("_", "")[:3].upper())
            for (a, p) in zip(attrnames, props)
            if isinstance(p, link_property)
        ]
        attrsvv = [a[::-1] for a in attrshierarch]
        return dict(attrshierarch + attrsvv)

    @classmethod
    def description(cls):
        """The description of the DataItem, parsed from docstring."""
        lines = [
            line[4:] if line.startswith("    ") else line
            for line in cls.__doc__.splitlines()
        ]
        return "\n".join(lines)

    @classmethod
    def description_oneline(cls):
        """Single line description, e.g. for DataItems table."""
        # This currently reconstructs one line descriptions from auto-generated
        # multi-line descriptions. Those auto-generated descriptions have
        # just the class name as the first line.
        # TODO: Do something nicer here.
        lines = cls.description().splitlines()
        if len(lines) > 2 and " " not in lines[0]:
            lines = lines[2:]
        return " ".join(lines)

    def __eq__(self, other):
        """Test equality by comparing persistent properties."""
        # Checking for identity adds the potential shortcut to keep
        # lazy lists lazy.
        if self is other:
            return True

        if type(self) != type(other):
            return False

        # TODO: Loading all the data all the time does not scale!
        # TODO: Perhaps just compare the filenames or so?
        if hasattr(self, "load_data"):
            self.load_data()
            other.load_data()

        # Ignore object_id when testing for equality.
        attrs_to_ignore = ["object_id"]
        attrnames = [a for a in self.get_persistent() if a not in attrs_to_ignore]
        for a in attrnames:
            v1 = getattr(self, a)
            if isinstance(v1, list):
                for _ in v1:
                    # Ensures a lazy list is instantiated.
                    pass
            v2 = getattr(other, a)
            if isinstance(v2, list):
                for _ in v2:
                    # Ensures a lazy list is instantiated.
                    pass

            if v1 != v2:
                return False

        return True

    def get_differences(self, other):
        """Show differences between two objects."""
        differences = ""
        for a in self.get_persistent():
            v1 = getattr(self, a)
            v2 = getattr(other, a)
            if v1 != v2:
                differences += f"{a} {v1} {v2}\n"
        return differences

    @classmethod
    def get_persistent_properties(cls, exclude_object_id=False):
        """Return a list of persistent attribute names.

        Like the common one in DBMeta.py, but preserving the order in which
        they are defined.
        TODO: Move to common DBMeta?
        """
        return [
            attr
            for attr, v in cls.__dict__.items()
            if isinstance(v, persistent_property)
            and not (exclude_object_id and attr == "object_id")
        ]

    def load_data(self):
        """Load data from FITS file"""
        filename = self.filename

        if not filename:
            return

        # A hash of the filename is linked to the filename
        # so the filename can be used in fits headers.
        # TODO: This is probably not true anymore and can be removed?
        if os.path.islink(filename) and not filename.endswith(".fits"):
            filename = os.readlink(filename)

        # Retrieve the file if it does not exist yet.
        if not os.path.exists(filename):
            self.retrieve()

        # Assume it's a fits file.
        with fits.open(filename) as hdulist:
            hdus: List[fits.ImageHDU] = [
                hdu for hdu in hdulist[1:] if isinstance(hdu, fits.ImageHDU)
            ]
            # "DET3.DATA" -> ('3', 'DATA')
            pattern_extname = re.compile(r"^DET(\d)\.(\w+)$")
            for hdu in hdus:
                assert "EXTNAME" in hdu.header, "Cannot find EXTNAME"
                extname = hdu.header["EXTNAME"]
                assert re.match(pattern_extname, extname), f"Bad EXTNAME {extname}"
            hdusd = {
                re.match(pattern_extname, hdu.header["EXTNAME"]).groups(): hdu
                for hdu in hdus
            }
            extensionnumbers = sorted(set(k for k, _ in hdusd))
            # E.g. ("1", "2", ..., "9")
            if hasattr(self, "extensions"):
                # prop = getattr(self, "extensions")
                for extensionnr, extension in zip(extensionnumbers, self.extensions):
                    hdudata = hdusd[(extensionnr, "DATA")]
                    # TODO: Compare headers of extension?
                    # extension_loaded = prop.prop_type(header=hdudata.header)
                    extension.image = hdudata.data
                    if (extensionnr, "ERR") in hdusd:
                        extension.uncertainty = hdusd[(extensionnr, "ERR")].data
                    if (extensionnr, "DQ") in hdusd:
                        extension.quality = hdusd[(extensionnr, "DQ")].data
            else:
                raise NotImplementedError(
                    "FITS files without extensions are not yet supported."
                )
