# -*- coding: utf-8 -*-
"""Test creation of associationmatrix."""
from pathlib import Path

from micado.toolbox.ditaacreator import DataRaw

from micado.toolbox.create_associationmatrix import (
    CLASSES_STANDARD_IMAGING,
    CLASSES_HIGH_CONTRAST_IMAGING,
    can_zig_zag,
    ConnPosArt,
    ConnCanvas,
    NeighbourArt,
    ContainedInt,
    CLASSES_STANDARD_IMAGING_FINAL,
    CLASSES_STANDARD_IMAGING_ALL,
    CLASSES_HIGH_CONTRAST_IMAGING_ALL,
    AssociationMatrix,
)


def test_matrix():
    """Test creation of the matrix."""
    path_associationmatrix = Path(__file__).resolve().parent / "associationmatrix.ditaa"
    with open(path_associationmatrix, "r") as f:
        matrix_expected = f.read()

    matrix = AssociationMatrix(
        CLASSES_STANDARD_IMAGING,
        CLASSES_STANDARD_IMAGING_FINAL,
        CLASSES_STANDARD_IMAGING_ALL,
    )
    matrix_generated = matrix.generate_association_matrix()
    assert matrix_generated.strip() == matrix_expected.strip()


def test_matrix_hci():
    """Test creation of the matrix for high contrast imaging."""
    path_associationmatrix = (
        Path(__file__).resolve().parent / "associationmatrix_hci.ditaa"
    )
    with open(path_associationmatrix, "r") as f:
        matrix_expected = f.read()

    matrix = AssociationMatrix(
        CLASSES_HIGH_CONTRAST_IMAGING, [], CLASSES_HIGH_CONTRAST_IMAGING_ALL
    )
    matrix_generated = matrix.generate_association_matrix()
    assert matrix_generated.strip() == matrix_expected.strip()


def test_shuffle():
    """Test whether the shuffle runs, it is not actually used.

    TODO: FIX test
    """
    matrix = AssociationMatrix(
        CLASSES_STANDARD_IMAGING, [], CLASSES_STANDARD_IMAGING_ALL
    )
    matrix.build_association_matrix()
    matrix.conncanvas.shuffle()
    matrix.shuffle(matrix.conncanvas)


def test_zig_zag():
    """Test zig zag algorithm."""
    assert can_zig_zag(
        "   xx   ",
        "xx     x",
    )

    assert can_zig_zag(
        "   x   ",
        "xx   xx",
    )

    assert not can_zig_zag(
        "   xx   ",
        "xxx    x",
    )

    assert not can_zig_zag(
        "     x  ",
        "xx   x  ",
    )

    assert not can_zig_zag(
        "   x  ",
        "  x   ",
    )

    assert can_zig_zag(
        "  x x  ",
        "       ",
    )


def test_overlap():
    """Test overlapping art."""
    poraw1 = ConnPosArt(
        x=-1,
        y=-1,
        art=DataRaw(text="DataRaw1"),
        centered=False,
    )
    poraw2 = ConnPosArt(
        x=3,
        y=1,
        art=DataRaw(text="DataRaw2"),
        centered=False,
        left=NeighbourArt(art=poraw1, distance=2),
    )
    poraw1.right = NeighbourArt(art=poraw2, distance=2)
    c = ConnCanvas(arts=[poraw1, poraw2])
    assert c.tension
    assert poraw2.tension_left
    for _ in range(20):
        c.shuffle()


def test_move_same():
    """Test moves with the same tension in main shuffle."""
    poraw1 = ConnPosArt(
        x=-1,
        y=-1,
        art=DataRaw(text="DataRaw1"),
        centered=False,
    )
    c = ConnCanvas(arts=[poraw1])
    for _ in range(10):
        c.shuffle()
    assert c.tension < 1


def test_move_same_other_shuffle():
    """Test moves with the same tension in other shuffle."""
    ciy = ContainedInt(4)
    cids = {
        "y": ciy,
    }
    poraw1 = ConnPosArt(
        x=2,
        y=ciy,
        art=DataRaw(text="DataRaw1"),
        centered=False,
    )
    c = ConnCanvas(arts=[poraw1])
    matrix = AssociationMatrix([], [], [])
    matrix.conncanvas = c
    matrix.contained_ints_dict = cids
    for _ in range(10):
        matrix.shuffle(c)

    print(c.tension)
