# -*- coding: utf-8 -*-
"""Tests for MasterDark.py."""
import os

from astropy.io import fits

from micado.config.Environment import Env
from micado.hardware.LinearityCurve import default_linearitycurve
from micado.hardware.QuantumEfficiencyCurve import default_quantumefficiencycurve
from micado.hardware.ShotNoise import ShotNoise
from micado.hardware.SummedExposure import SummedExposure
from micado.main.Img import Img
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.Dark import Dark
from micado.dataitems.Dark import DarkExtension
from micado.dataitems.MasterDark import MasterDark
from micado.main.Imstat import Imstat
from micado.requirements.DPSmasterdark import DPSmasterdark
from micado.templates.MicadoImgCalDark import MicadoImgCalDark
from micado.recipes.McdDetDark import McdDetDark
from micado.test.functional.dark_test import remove_file_and_links
from micado.util.esokeywords import verify_fitsheaders


def create_masterdark_input(simple=False, make=True, ndetectors=1):
    """Create input for MasterDark for testing.

    TODO: Replace with fixture?
    """
    # Create a MasterDark with 1 dark of 16x16 pixels when simple is set.
    n_darks = 5 if not simple else 1
    # TODO: Put 4096 back at some point
    size = 32

    mydetector = Detector(
        x_size=size,
        y_size=size,
    )
    mydetectorlist = [mydetector] * ndetectors

    # Set the real dark current and create a DetectorArray.
    # The default dark current is 0.1 e/s. Setting the dark current too high
    # will cause interference with the other effects, making it it difficult
    # to recover.
    dark_real = 1.0
    mydarkcurrent = DarkCurrent(value=dark_real)

    myshotnoise = ShotNoise(state=1)
    mydetector = DetectorArray(
        name="",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        # TODO: disable qe_curve and linearity, set include=False or so
        #       but that won't work when reading them in.
        qe_curve=default_quantumefficiencycurve,
        linearity=default_linearitycurve,
        exposure_action=SummedExposure(state=1),
        readout_noise=None,
    )

    # A template
    dark_template = MicadoImgCalDark(det__dit=1, det__ndit=1)

    # An observingblock
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    # Create raw Darks.
    mydarks = []
    for _ in range(n_darks):
        # DarkExtensions
        myextensions = [
            DarkExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
        ]

        mydark = Dark(
            detector=mydetector,
            template=dark_template,
            observing_block=dark_ob,
            extensions=myextensions,
        )
        if make:
            mydark.make_simulation()
            mydark.save_fits()

        mydarks.append(mydark)

    return dark_real, mydarks


def create_masterdark(simple=False, make=True, ndetectors=1):
    """Create MasterDark for testing."""
    dark_real, mydarks = create_masterdark_input(
        simple=simple, make=make, ndetectors=ndetectors
    )

    # A recipe. The dark recipe has no parameters.
    myrecipe = McdDetDark()

    # Create MasterDark.
    mymasterdark = MasterDark(
        recipe=myrecipe,
        darks=mydarks,
    )
    if make:
        mymasterdark.make_proto()
        mymasterdark.save_fits()
    return mymasterdark, dark_real, mydarks


def compare_pros(dataitem1: Img, dataitem2: Img):
    """Compare two processed Img DataItems.

    Currently works for MasterDark and MasterFlat.
    """
    assert dataitem1.fits_keywords == dataitem2.fits_keywords
    assert hasattr(dataitem1, "recipe")
    assert hasattr(dataitem2, "recipe")
    # noinspection PyUnresolvedReferences
    assert dataitem1.recipe == dataitem2.recipe

    ext1, ext2 = dataitem1.extensions[0], dataitem2.extensions[0]
    assert ext1.fits_keywords == ext2.fits_keywords
    assert ext1 == ext2, ext1.get_differences(ext2)

    # TODO: do something generic here
    for attr in ["darks", "flats"]:
        if hasattr(dataitem1, attr):
            for d1, d2 in zip(getattr(dataitem1, attr), getattr(dataitem2, attr)):
                assert d1.fits_keywords == d2.fits_keywords
                assert d1 == d2, d1.get_differences(d2)

    assert dataitem1 == dataitem2


def test_processing():
    """Test creation of MasterDark."""
    mymasterdark, dark_real, mydarks_raw = create_masterdark()

    # Create image statistics.
    myimstatraw = Imstat(baseframe=mydarks_raw[0])
    myimstatraw.make()

    myimstat = Imstat(baseframe=mymasterdark)
    myimstat.make()

    # The standard deviation scales with sqrt(n_raws).
    # This check will fail if the the darks are correlated, e.g. if something
    # goes wrong with the random number generation. However, it can always
    # fail because this is a stochastic process.
    assert myimstat.std < myimstatraw.std / 1.5

    # Verify that the measured dark current is the same as the real current.
    myreq = DPSmasterdark()
    myreq.precision = 0.05
    myreq.dark_true = dark_real
    myreq.masterdarkqc = mymasterdark.qc
    myreq.make()
    assert myreq.passed, "Requirement not met!"

    myreqfail = DPSmasterdark()
    myreqfail.precision = 0.05
    myreqfail.dark_true = dark_real * 8
    myreqfail.masterdarkqc = mymasterdark.qc
    myreqfail.make()
    assert not myreqfail.passed, "Requirement met inadvertently."

    fn = mymasterdark.filename
    verify_fitsheaders(fn)

    master2 = MasterDark(filename=fn)
    compare_pros(mymasterdark, master2)

    for dm in mydarks_raw:
        remove_file_and_links(dm.filename)

    remove_file_and_links(mymasterdark.filename)


def test_none_mobject_fits():
    """Test an '' reference in a FITS file."""

    masterdark1, _, mydarks_raw = create_masterdark(simple=True)

    # Replace 'HIERARCH ESO REC PROCESS' = None with ''
    hdus = fits.open(os.path.realpath(masterdark1.filename), mode="update")
    hdus[0].header["HIERARCH ESO REC PROCESS"] = ""
    hdus.close()

    masterdark2 = MasterDark(filename=masterdark1.filename)
    assert masterdark1 == masterdark2

    # Cleanup.
    for dm in mydarks_raw:
        remove_file_and_links(dm.filename)

    remove_file_and_links(masterdark1.filename)
