# -*- coding: utf-8 -*-
"""Test ScopeSIM installation.

Based on
https://github.com/astronomyk/ScopeSim/blob/master/scopesim/tests/tests_integrations/test_simplecado.py
"""

import numpy
import scopesim
import scopesim_templates
from astropy.io import fits

from micado.hardware.DarkCurrent import DarkCurrent
from micado.test.functional.science_test import get_science

# TODO: OBS_DIT and OBS_NDIT are renamed in newer ScopeSIM versions.
OBS_DIT = "!OBS.dit"
OBS_NDIT = "!OBS.ndit"

# DETECTOR nested dictionary
# Ideally I would like to have "kwargs" : {"filename" : "MICADO_detectors.tbl"}
# so that MICADO_detectors.tbl is the one place where the detector description
# is stored. This however is more suited to storing the info in a database
DETECTOR_YAML = {
    "object": "detector",
    "alias": "DET",
    "name": "test_detector",
    "properties": {"dit": OBS_DIT, "image_plane_id": 0},
    "effects": [
        {
            # TODO: detector_array_list is renamed in IRDB.
            "name": "detector_array_list",
            "description": "SimpleCADO detector array list",
            "class": "DetectorList",
            "kwargs": {
                "array_dict": {
                    "id": [1],
                    "pixel_size": [0.015],
                    "angle": [0.0],
                    "gain": [1.0],
                    "x_cen": [0],
                    "y_cen": [0],
                    "x_size": [4096],
                    "y_size": [4096],
                },
                # These units are the default, so they can also be removed.
                "x_cen_unit": "mm",
                "y_cen_unit": "mm",
                "x_size_unit": "pix",
                "y_size_unit": "pix",
                "pixel_size_unit": "mm",
                "angle_unit": "deg",
                "gain_unit": "electron/adu",
            },
        },
        {
            "name": "dark_current",
            "description": "SimpleCADO dark current",
            "class": "DarkCurrent",
            # [e-/s] level of dark currentSimpleCADO.yaml
            "kwargs": {"value": 0.2, "dit": OBS_DIT, "ndit": OBS_NDIT},
        },
    ],
}


OBSERVATIONS_DICT = {
    # Not yet implemented
    OBS_NDIT: 1,
    OBS_DIT: 10,
    # because optical train still need these
    "!INST.pixel_scale": 0.004,
    "!INST.plate_scale": 0.2666667,
}


def test_simplecado():
    """Test SimpleCADO, just a Dark."""
    # By using empty_sky it is ensured that we only measure the instrument.
    src = scopesim.source.source_templates.empty_sky()

    cmd = scopesim.commands.UserCommands(
        yamls=[DETECTOR_YAML],
        properties=OBSERVATIONS_DICT,
    )

    opt = scopesim.OpticalTrain(cmd)
    opt.observe(src)
    hdulists = opt.readout()
    # There can be multiple hdulists in case observations are done with
    # multiple instruments at the same time.
    hdulist = hdulists[0]
    # The first HDU never contains an image, even if there is only one.
    hdu = hdulist[1]

    # Finished - now just testing the output
    # No rogue photon torpedoes
    print(opt.image_planes[0].image)
    assert numpy.all(opt.image_planes[0].image == 0)

    # dark = 0.2 ct/s, DIT = 10s, NDIT = 1
    print(hdu.data)
    assert numpy.all(hdu.data == 2.0)


def test_headers():
    """Test whether the headers of MicadoWISE are the same as ScopeSIM"""
    # Explicitely use defaults.
    science1 = get_science(
        size=1024,
        dark_current=DarkCurrent.value.default,
    )
    science1.observing_block.target = "cluster"

    # Get the ScopeSIM input.
    observations_dict = science1.get_observations_dict()
    # TODO: Shift source?
    target = science1.observing_block.target

    science1.make_simulation()
    science1.save_fits()
    fn1 = science1.filename

    fn2 = "MCD_direct_" + fn1
    src_function = getattr(scopesim_templates.micado, target)
    src = src_function()

    cmd = scopesim.UserCommands(use_instrument="MICADO", set_modes=["SCAO", "IMG_4mas"])
    cmd.update(properties=observations_dict)

    micado = scopesim.OpticalTrain(cmd)
    micado.observe(src)
    micado.readout(filename=fn2)

    bad_keys_all = {}
    with fits.open(fn1) as hdus1, fits.open(fn2) as hdus2:
        assert len(hdus1) == len(hdus2)
        for i, (hdu1, hdu2) in enumerate(zip(hdus1, hdus2)):
            headers1 = hdu1.header
            headers2 = hdu2.header
            bad_keys = {}

            for k in headers1:
                v1 = headers1[k]
                v2 = headers2[k] if k in headers2 else None
                if v1 != v2:
                    bad_keys[k] = (v1, v2)

            lbks = list(bad_keys.items())
            if lbks:
                bad_keys_all[i] = lbks
    assert not bad_keys_all, bad_keys_all
