# -*- coding: utf-8 -*-
"""Tests whether CPL plugin works."""
import pytest
from micado.recipes.McdDetDepersist import McdDetDepersist

from micado.dataitems.DepersistedDark import DepersistedDark
from micado.dataitems.MasterDark import MasterDark
from micado.recipes.McdDetDark import McdDetDark
from micado.test.functional.dark_test import generate_dark
from micado.test.functional.masterdark_test import create_masterdark


def compare_fits_keywords(keywords1, keywords2):
    """Compare FITS keywords."""
    keywords_both = [k for k in keywords1 if k in keywords2]
    keywords_just1 = [k for k in keywords1 if k not in keywords_both]
    keywords_just2 = [k for k in keywords2 if k not in keywords_both]
    keywords_different = [k for k in keywords_both if keywords1[k] != keywords2[k]]
    return keywords_just1, keywords_just2, keywords_different


@pytest.mark.skip(reason="The CPL Recipe only works with depersisted darks.")
def test_cpl_processing():
    """Test whether CPL and prototype processing is identical."""
    # TODO: do something smarter than marking the test as xfail, e.g.
    #       only run the test when the MICADO Pipeline is installed.

    (mymasterdark_proto, dark_real, mydarks_raw,) = create_masterdark(
        make=True,
        ndetectors=9,
    )

    # A recipe. The dark recipe has no parameters.
    myrecipe = McdDetDark()

    mymasterdark_cpl = MasterDark(
        recipe=myrecipe,
        darks=mydarks_raw,
    )

    mymasterdark_proto.make_proto()

    mymasterdark_cpl.make_cpl()

    for extension_proto, extension_cpl in zip(
        mymasterdark_proto.extensions,
        mymasterdark_cpl.extensions,
    ):
        # It seems that cpl is 32 bit and the prototype is 64 bit.
        # TODO: Verify that cpl should indeed be 32 bit.
        assert (
            extension_proto.image.astype(extension_cpl.image.dtype)
            == extension_cpl.image
        ).all()
        uncertainty_ratio = extension_cpl.uncertainty / extension_proto.uncertainty
        # uncertainty_ratio is about 2.8 for all pixels
        assert (0.8 < uncertainty_ratio).all()
        assert (uncertainty_ratio < 3.0).all()

    keywords_proto, keywords_cpl, keywords_different = compare_fits_keywords(
        mymasterdark_proto.fits_keywords, mymasterdark_cpl.fits_keywords
    )
    print(keywords_proto)
    print(keywords_cpl)
    print(keywords_different)
    # TODO: Ensure that the recipes and the prototypes produce the same
    #       filenames.
    assert all(
        k.startswith("HIERARCH ESO PRO REC") and k.endswith(" NAME")
        for k in keywords_different
    )
    assert not keywords_cpl
    # TODO: Support QC
    assert all("QC" in k for k in keywords_proto)


@pytest.mark.skip(reason="The CPL Recipe only works with depersisted darks.")
def test_loading_cpl_data():
    """Test whether loading and saving data results in the same dataitem."""
    mymasterdark_proto, dark_real, mydarks_raw = create_masterdark(
        make=False,
        ndetectors=9,
    )

    # A recipe. The dark recipe has no parameters.
    myrecipe = McdDetDark()

    mymasterdark_cpl = MasterDark(
        recipe=myrecipe,
        darks=mydarks_raw,
    )
    mymasterdark_cpl.make_cpl()

    fn_test = "MCDWS_testmasterdark.fits"
    mymasterdark_cpl.result.IMG_MASTERDARK.writeto(fn_test, overwrite=True)

    mymasterdark_read = MasterDark(filename=fn_test)
    # TODO: Ensure QC works
    mymasterdark_read.qc = None
    # Hack because filename is not generated
    mymasterdark_cpl.filename = fn_test
    assert mymasterdark_cpl == mymasterdark_read


def test_cpl_with_depersist():
    """Test CPL with depersist recipe."""
    mydark = generate_dark(n_det=9)
    mydark.save_fits()

    myrecipe = McdDetDepersist()
    mydepersisteddark = DepersistedDark(
        raw=mydark,
        recipe=myrecipe,
    )
    mydepersisteddark.make_cpl()
    for ext_raw, ext_dp in zip(
        mydark.extensions,
        mydepersisteddark.extensions,
    ):
        assert (ext_dp.image == ext_raw.image.astype("float32")).all()
        assert (ext_dp.quality == 0).all()
        assert len(mydepersisteddark.extensions) == 9
