# -*- coding: utf-8 -*-
"""Augment classes for backward chaining.

TODO: Put the lineage somewhere in the data model.
"""
import operator
from typing import Type

from micado.dataitems.Astrom import Astrom
from micado.dataitems.Background import Background
from micado.dataitems.Calibrated import Calibrated
from micado.dataitems.Dark import Dark
from micado.dataitems.Detrend import Detrend
from micado.dataitems.DetrendAi import DetrendAi
from micado.dataitems.DetrendAstromfield import DetrendAstromfield
from micado.dataitems.DetrendBase import DetrendBase
from micado.dataitems.DetrendBkgfield import DetrendBkgfield
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.DetrendFinalAi import DetrendFinalAi
from micado.dataitems.DetrendFinalAstromfield import DetrendFinalAstromfield
from micado.dataitems.DetrendFinalBase import DetrendFinalBase
from micado.dataitems.DetrendFinalStdfield import DetrendFinalStdfield
from micado.dataitems.DetrendFinalWam import DetrendFinalWam
from micado.dataitems.DetrendIllum import DetrendIllum
from micado.dataitems.DetrendStdfield import DetrendStdfield
from micado.dataitems.DetrendWam import DetrendWam
from micado.dataitems.DistortionElt import DistortionElt
from micado.dataitems.DistortionWam import DistortionWam
from micado.dataitems.Flat import Flat
from micado.dataitems.IllumCorr import IllumCorr
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Nonlin import Nonlin
from micado.dataitems.Nonlinearity import Nonlinearity
from micado.dataitems.Persistence import Persistence
from micado.dataitems.PhotomScience import PhotomScience
from micado.dataitems.PhotomStdfield import PhotomStdfield
from micado.dataitems.PhotomBase import PhotomBase
from micado.dataitems.Stack import Stack
from micado.main.Raw import Raw


def augment_backward_chaining_detrend(cls_detrend: Type[DetrendBase]):
    """Augment detrend class with backward chaining."""
    cls_detrend.lineage = {
        cls_detrend.masterflat: [
            (cls_detrend.det_dit, operator.eq, MasterFlat.det_dit),
            (cls_detrend.filtname, operator.eq, MasterFlat.filtname),
            (cls_detrend.readmode, operator.eq, MasterFlat.readmode),
        ],
        cls_detrend.masterdark: [
            (cls_detrend.det_dit, operator.eq, MasterDark.det_dit),
            (cls_detrend.readmode, operator.eq, MasterDark.readmode),
        ],
        cls_detrend.raw: [
            (
                cls_detrend.mjdobs,
                operator.eq,
                cls_detrend.get_progenitor_class().mjdobs,
            ),
        ],
        cls_detrend.nonlinearity: [],
    }


def augment_backward_chaining_detrendfinal(cls_detrendfinal: Type[DetrendFinalBase]):
    """Augment DetrendFinal dataitems with backward chaining information."""
    cls_detrendfinal.lineage = {
        cls_detrendfinal.detrend: [
            (
                cls_detrendfinal.mjdobs,
                operator.eq,
                cls_detrendfinal.get_progenitor_class().mjdobs,
            )
        ],
        cls_detrendfinal.background: [
            (cls_detrendfinal.tpl__start, operator.eq, Background.tpl__start),
        ],
        cls_detrendfinal.illum: [
            (cls_detrendfinal.filtname, operator.eq, IllumCorr.filtname)
        ],
    }


def augment_backward_chaining_photom(cls_photom: Type[PhotomBase]):
    """Augment PhotomBase dataitems with backward chaining information."""
    cls_photom.lineage = {
        cls_photom.detrend: [
            (cls_photom.mjdobs, operator.eq, cls_photom.get_progenitor_class().mjdobs)
        ],
        cls_photom.astrom: [(cls_photom.mjdobs, operator.eq, Astrom.mjdobs)],
        cls_photom.photomref: [],
    }


def augment_backward_chaining():
    """Augment dataitems with backward chaining information."""
    Stack.lineage = {
        Stack.detrends: [(Stack.tpl__start, operator.eq, Calibrated.tpl__start)]
    }

    Calibrated.lineage = {
        Calibrated.detrend: [(Calibrated.mjdobs, operator.eq, DetrendFinal.mjdobs)],
        Calibrated.astrom: [(Calibrated.mjdobs, operator.eq, Astrom.mjdobs)],
        # TODO: Add selection mechanism for PhotomStdfield or PhotomScience.
        # Selecting on (Calibrated.mjdobs, operator.eq, PhotomBase.mjdobs) would
        # only select PhotomScience, because only those have the same MJDOBS.
        # Instead, select on DIT and Filter name, similar to MasterFlat.
        # This will always prefer the PhotomScience if available though.
        Calibrated.photom: [
            (Calibrated.det_dit, operator.eq, PhotomBase.det_dit),
            (Calibrated.filtname, operator.eq, PhotomBase.filtname),
        ],
    }

    # PhotomBase itself does not have a lineage, since it is a base class.
    augment_backward_chaining_photom(PhotomScience)
    augment_backward_chaining_photom(PhotomStdfield)

    Astrom.lineage = {
        Astrom.detrend: [(Astrom.mjdobs, operator.eq, DetrendFinal.mjdobs)],
        Astrom.astromref: [],
        # TODO: Something for these when Astrometric imaging is implemented.
        Astrom.distortion_wam: [],
        Astrom.distortion_elt: [],
    }

    # TODO: Something for these when Astrometric imaging is implemented.
    DistortionWam.lineage = {
        DistortionWam.detrend: [
            # TODO: Fix this incorrect dummy.
            (
                DistortionWam.instname,
                operator.eq,
                DetrendFinal.instname,
            ),
        ],
    }

    # TODO: Something for these when Astrometric imaging is implemented.
    DistortionElt.lineage = {
        DistortionElt.detrend: [
            # TODO: Fix this incorrect dummy.
            (
                DistortionElt.instname,
                operator.eq,
                DetrendFinal.instname,
            ),
        ],
    }

    Background.lineage = {
        Background.detrends: [
            (Background.tpl__start, operator.eq, DetrendBkgfield.tpl__start),
        ],
    }

    IllumCorr.lineage = {
        IllumCorr.detrends: [
            # TODO: Does the Illumination Correction depend on exposure time?
            # (IllumCorr.det_dit, operator.eq, Detrend.det_dit),
            (IllumCorr.filtname, operator.eq, DetrendIllum.filtname),
        ],
        # The photometric reference catalog is selected only on CATG.
        IllumCorr.photomref: [],
    }

    augment_backward_chaining_detrend(Detrend)
    augment_backward_chaining_detrend(DetrendAi)
    augment_backward_chaining_detrend(DetrendIllum)
    augment_backward_chaining_detrend(DetrendStdfield)
    augment_backward_chaining_detrend(DetrendBkgfield)
    augment_backward_chaining_detrend(DetrendWam)
    augment_backward_chaining_detrend(DetrendAstromfield)

    augment_backward_chaining_detrendfinal(DetrendFinal)
    augment_backward_chaining_detrendfinal(DetrendFinalAi)
    augment_backward_chaining_detrendfinal(DetrendFinalStdfield)
    augment_backward_chaining_detrendfinal(DetrendFinalWam)
    augment_backward_chaining_detrendfinal(DetrendFinalAstromfield)

    Nonlinearity.lineage = {
        Nonlinearity.flats: [
            # (Nonlinearity.instname, operator.eq, Flat.instname),
            # (Nonlinearity.readmode, operator.eq, Flat.readmode),
            (Nonlinearity.readmode, operator.eq, Nonlin.readmode),
        ],
    }

    MasterFlat.lineage = {
        MasterFlat.flats: [
            (MasterFlat.det_dit, operator.eq, Flat.det_dit),
            (MasterFlat.readmode, operator.eq, Flat.readmode),
            (MasterFlat.filtname, operator.eq, Flat.filtname),
        ],
        MasterFlat.masterdark: [
            (MasterFlat.det_dit, operator.eq, MasterDark.det_dit),
            (MasterFlat.readmode, operator.eq, MasterDark.readmode),
        ],
    }

    MasterDark.lineage = {
        MasterDark.darks: [
            (MasterDark.det_dit, operator.eq, Dark.det_dit),
            (MasterDark.readmode, operator.eq, Dark.readmode),
        ],
    }

    Persistence.lineage = {
        Persistence.raw: [(Persistence.mjdobs, operator.eq, Raw.mjdobs)],
        Persistence.raws: [(Persistence.mjdobs, operator.gt, Raw.mjdobs)],
    }
