# -*- coding: utf-8 -*-
"""Generate .ditaa figures from DataItems."""

from collections import OrderedDict
from typing import Type, List

from micado.dataitems.AstromRef import AstromRef
from micado.dataitems.DetrendFinalPupil import DetrendFinalPupil
from micado.dataitems.DetrendFinalStdfield import DetrendFinalStdfield
from micado.dataitems.DetrendFinalWaffle import DetrendFinalWaffle
from micado.dataitems.DpsicsOptSciBase import DpsicsOptSciBase
from micado.dataitems.PhotomRef import PhotomRef
from micado.dataitems.Science import Science
from micado.dataitems.ScienceCoroPupil import ScienceCoroPupil
from micado.dataitems.ScienceCoroWaffle import ScienceCoroWaffle
from micado.dataitems.Stdfield import Stdfield
from micado.main.Cat import Cat
from micado.main.MDBObjectMeta import MDBObjectMeta
from micado.main.Template import Template
from micado.recipes.DerivePersistence import DerivePersistence
from micado.templates.McdImgCalQacits45Base import McdImgCalQacits45Base
from micado.templates.McdImgObsSiaiBase import McdImgObsSiaiBase
from micado.templates.MicadoCoroAcqStar import MicadoCoroAcqStar
from micado.templates.MicadoCoroObsFpm import MicadoCoroObsFpm
from micado.templates.MicadoImgObsBgVirtual import MicadoImgObsBgVirtual
from micado.templates.MicadoImgObs import MicadoImgObs
from micado.templates.McdImgSciObjectHciBase import McdImgSciObjectHciBase

from micado.toolbox.ditaacreator import (
    SimpleStepInput,
    SimpleStep,
    SimpleRecipe,
    DataRaw,
    DataFinal,
    DataIntermediate,
    DataStatic,
    RecipeStep as RecipeStepDitaa,
    RecipeFull,
    Process,
    RecipeBox,
    Template as TemplateDitaa,
    SimpleStepParams,
    ParamsBox,
    Data,
)

from micado.main.RecipeStepExtra import RecipeStepBase
from micado.main.Img import Img
from micado.main.Tab import Tab
from micado.main.DataItem import DataItem
from micado.main.Raw import Raw
from micado.main.Ext import Ext
from micado.main.Final import Final
from micado.main.StaticTable import StaticTable
from micado.util.derivedclasses import get_dataitems_derived, get_raws_derived
from micado.util.paths import PATH_DOCS_GENERATED

from micado.dataitems.CalibHCI import CalibHCI
from micado.dataitems.Calibrated import Calibrated
from micado.dataitems.CubeHCI import CubeHCI
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.PhotomBase import PhotomBase
from micado.dataitems.Stack import Stack

from micado.util.classlists import (
    CLASSES_DATAITEMS,
    DetrendHCIBase,
    DetrendBase,
)


def get_ditaaclass(attrtype, final=None) -> Type[Data]:
    """Get the ditaa class for a given MicadoWISE class."""
    final = final or (
        Calibrated,
        Stack,
        CalibHCI,
        CubeHCI,
    )
    if issubclass(attrtype, tuple(final)):
        return DataFinal
    ditaaclass_from_wiseclass = OrderedDict(
        [
            (Raw, DataRaw),
            (Final, DataFinal if not final else DataIntermediate),
            # TODO: Fix External data, now they are just hardcoded.
            (Ext, DataStatic),
            (AstromRef, DataStatic),
            (PhotomRef, DataStatic),
            (Img, DataIntermediate),
            (DataItem, DataIntermediate),
        ]
    )
    # # Overwrite
    # if external:
    #     # Remove things incorrectly External
    #     for clsn, clsd in ditaaclass_from_wiseclass.items():
    #         if clsn not in external and clsd == DataExternal:
    #             ditaaclass_from_wiseclass[clsn] = DataIntermediate
    #     for clsn in external:
    #         if clsn in ditaaclass_from_wiseclass:
    #             ditaaclass_from_wiseclass.pop(clsn)
    #
    #         [cls] = DataFinal

    for clsi, clsd in ditaaclass_from_wiseclass.items():
        if issubclass(attrtype, clsi):
            clsn = clsd
            break
    else:
        clsn = DataIntermediate
    return clsn


def get_step(mo: RecipeStepBase) -> SimpleStep:
    """Create a ditaa SimpleStep from a WISE class."""
    prev = mo.get_previous()
    inputs = [
        SimpleStepInput(
            cls=get_ditaaclass(attrtype),
            text=attrtype.name_eso(),
        )
        for attrtype in mo.get_other_inputdata_full()
        if attrtype != prev
    ]
    params = [pn.upper() for pn in mo.get_parameter_names()]

    allowed_raw = None
    if mo == DerivePersistence:
        allowed_raw = [Raw]

    return SimpleStep(
        process=mo.name_eso(),
        inputs=inputs,
        params=SimpleStepParams(text=params) if len(params) else None,
        # TODO: so this is not necessarily always a raw, resulting in the wrong
        #       color later on. Fix this.
        raw=DataRaw(text_from_dataitem(prev, allowed=allowed_raw, strict=True))
        if not issubclass(prev, RecipeStepBase)
        else None,
        previous=get_step(mo.get_previous_step())
        if mo.get_previous_step() is not None
        else None,
        filename=mo.__name__,
    )


def text_from_template(template: Template, allowed=None):
    """Get text for a ditaa box for a Template."""
    # Some Recipes (e.g. Persistence) allow Raws from multiple templates.
    # TODO: UnspersistedSky should not allow e.g. MCD_img_cal_dark.
    #       But those figures are not included in the documentation anyway.
    temps_derived = [
        temp
        for temp in template.get_derived_classes()
        if "qacits" not in temp.name_eso() and not temp.name_eso().endswith("base")
        # TODO: Find better way to include/exclude DPSICS things.
        and temp not in [MicadoCoroAcqStar]
    ]
    if allowed is not None:
        temps_derived = [temp for temp in temps_derived if temp in allowed]
    # Do some sanity checks.
    assert not (
        template == Template and not temps_derived
    ), "Derived templates are required if base template is used."
    # A sanity check that
    assert not (
        # TODO: Figure out whether this works as desired for MicadoImgObs
        template
        not in (
            Template,
            MicadoImgObs,
            McdImgObsSiaiBase,
            McdImgSciObjectHciBase,
            McdImgCalQacits45Base,
            # Also does background image:
            MicadoCoroObsFpm,
            # Hack to create backgrounds from different templates:
            MicadoImgObsBgVirtual,
        )
        and temps_derived
    ), f"Derived templates not supported for {template.name_eso()}"
    is_base_template = (template == Template) or template.__name__.endswith("Base")
    if is_base_template:
        temps = temps_derived
    elif allowed is not None and template not in allowed:
        temps = temps_derived
    else:
        temps = [template] + temps_derived

    assert temps, (template, temps_derived, template in allowed)
    # TODO: figure out a better way to sort this list.
    temps.sort(key=lambda t: t.name_eso())
    return "\n".join(tempi.name_eso() for tempi in temps)


def text_from_dataitem(dataitem: Type[DataItem], allowed=None, strict=False):
    """Get text for a ditaa box for a DataItem."""
    dataitems_derived1 = get_dataitems_derived(dataitem, allowed=allowed, strict=strict)
    dataitems_derived = [
        di
        for di in dataitems_derived1
        # TODO: Find better way to include/exclude DPSICS things.
        if not issubclass(di, DpsicsOptSciBase)
    ]
    return " or\n".join(di.name_eso() for di in dataitems_derived)


def ditaa_from_wise(mo: MDBObjectMeta):
    """Create a ditaa file from a DataItem class."""
    # Assert necessary to placate type checker.
    assert isinstance(mo, MDBObjectMeta)

    serror = "ditaa_from_wise can only be ran on processed Img or Tab or StaticTable dataitems"
    assert issubclass(mo, (Img, StaticTable, Tab, Cat)), serror

    recipe = mo.recipe.prop_type.name_eso()

    raw = mo.get_raw_class()
    assert issubclass(raw, Raw)

    products_derived = get_dataitems_derived(mo)
    progenitors_allowed = [pd.get_progenitor_class() for pd in products_derived]
    progenitor = mo.get_progenitor_class()
    progenitors_derived = get_dataitems_derived(
        progenitor,
        include_self=True,
        allowed=progenitors_allowed,
    )
    raws = get_raws_derived(mo)

    # TODO: Fix Photom hack.
    # It should be possible to derive a photometric solution also from
    # coronagraphic imaging. E.g. there should be PhotomXxxx classes
    # for that.
    if mo is PhotomBase:
        progenitors_derived = [
            DetrendFinalPupil,
            DetrendFinalStdfield,
            DetrendFinalWaffle,
            DetrendFinal,
        ]
        raws = [
            ScienceCoroPupil,
            Stdfield,
            ScienceCoroWaffle,
            Science,
        ]

    template = raw.template.prop_type
    # Ensure that e.g. raw darks cannot be detrended.
    templates_allowed = [template] + [ra.template.prop_type for ra in raws]
    text_template = text_from_template(template, allowed=templates_allowed)

    text_progenitor = text_from_dataitem(progenitor, allowed=progenitors_derived)
    text_product = text_from_dataitem(mo)

    steps: List[SimpleStep] = []
    # Assert necessary to placate type checker.
    assert issubclass(mo, (Img, StaticTable, Tab, Cat)), serror
    assert hasattr(mo, "recipe"), "Need a class with a recipe"
    mmo = mo.recipe.prop_type
    rs = mmo.process.prop_type
    while rs is not None:
        step = get_step(rs)
        steps.append(step)
        rs = rs.get_previous_step()

    clsp: Type[Data] = get_ditaaclass(mo)

    sr = SimpleRecipe(
        template=text_template,
        raw=text_progenitor,
        recipe=recipe,
        steps=steps[::-1],
        product=text_product,
    )

    asteps = [
        RecipeStepDitaa(
            process=Process(step.process),
            inputs=[n.cls(n.text) for n in step.inputs],
            params=ParamsBox(text=step.params.text)
            if step.params is not None
            else None,
            raw=step.raw if step.raw is not None else None,
            previous=Process(step.previous.process)
            if step.previous is not None
            else None,
            filename=step.filename,
        )
        for step in sr.steps
    ]

    # noinspection PyArgumentList
    # https://youtrack.jetbrains.com/issue/PY-36375
    product = clsp(sr.product)
    return RecipeFull(
        template=TemplateDitaa(sr.template) if sr.template is not None else None,
        raw=DataRaw(sr.raw),
        steps=asteps,
        recipe=RecipeBox(text=sr.recipe),
        product=product,
    )


def camel_from_snake(value):
    """Convert apply_master_dark back to ApplyMasterDark."""
    words = [word.capitalize() for word in value.split("_")]
    return "".join(words)


def get_applicable_templates_from_dataitems(dataitems):
    """Get applicable templates from dataitems."""
    templates = [
        raw.template.prop_type for raw in dataitems if hasattr(raw, "template")
    ]
    # Templates with Base in the name should never end up in figures.
    return [t for t in templates if not t.__name__.endswith("Base")]


def main():
    """Generate ditaa files for all revelant classes."""
    for cls in CLASSES_DATAITEMS:
        if cls.base_doc() in CLASSES_DATAITEMS:
            # TODO: Fix this hack
            if cls not in [DetrendHCIBase, DetrendBase]:
                continue
        fn = PATH_DOCS_GENERATED / f"{cls.__name__}.ditaa"
        rf = ditaa_from_wise(cls)
        s = str(rf)
        with open(fn, "w") as f:
            f.write(s)

        for rs in rf.steps:
            s = "\n".join(rs.lines_selfcontained)
            fn = PATH_DOCS_GENERATED / f"{rs.filename}.ditaa"
            with open(fn, "w") as f:
                f.write(s)


if __name__ == "__main__":
    main()
