# -*- coding: utf-8 -*-
"""Generate Template modules from .tsf files.

This file is not currently used. The template files are generated
from VODML instead. This file is used for sanity checks. It is not
clear in what format the actual templates will be delivered. This
file can be used again if that format is PAF.
"""

import glob
from collections import OrderedDict

from micado.util.esokeywords import esokeywords
from micado.util.paths import PATH_TEMPLATE_TSF

python_from_type = {
    "number": float,
    "integer": int,
    "keyword": str,
}

default_from_type = {
    "number": 0.0,
    "integer": 0,
    "keyword": "",
}


def parse_line(line: str):
    """Parse a line.

    Args:
        line: Line from a TSF file.
    """
    name = line.split()[0]
    l2 = line[len(name) :].strip()
    l2 += ";" if ";" not in l2 else ""
    v1, c1 = l2.split(";")
    value = v1.strip().strip('"').strip()
    comment = c1.strip(" #")
    return name, value, comment


def template_from_tsf(filename: str):
    """Create Template class from tsf file.

    Args:
        filename: Name of the TSF file to read.
    """
    lines1 = open(filename).readlines()
    lines2 = [line.strip() for line in lines1 if len(line.strip())]
    lines3 = [
        line
        for line in lines2
        if not line.startswith("#") and not line.startswith("PAF.HDR.")
    ]
    lines4 = [parse_line(line) for line in lines3]

    names_params = [line[1] for line in lines4 if line[0] == "TPL.PARAM"]
    lines5 = [
        (line[0], (line[1], line[2])) for line in lines4 if line[0] != "TPL.PARAM"
    ]

    lines6 = OrderedDict(lines5)
    params = OrderedDict(
        [
            (
                name,
                {
                    attr.lower(): lines6.get("{}.{}".format(name, attr), (None, ""))
                    for attr in [
                        "TYPE",
                        "RANGE",
                        "DEFAULT",
                        "VALUE",
                        "LABEL",
                        "MINIHELP",
                        "HIDE",
                    ]
                },
            )
            for name in names_params
        ]
    )

    name_template_eso = lines6["PAF.NAME"][0]
    name_template = "".join(
        a.lower().capitalize() for a in name_template_eso.split("_")
    )

    doc_template = lines6["PAF.DESC"][0]

    for name, param in params.items():
        name_attr_eso = name.replace(".", " ").upper()
        if name_attr_eso in esokeywords:
            esoattr = esokeywords.get(name_attr_eso)
            param["desc"] = esoattr.get("description").replace('"', "'")
            param["unit"] = esoattr.get("unit")
        else:
            param["desc"] = "None"
            param["unit"] = "None"

        if param["default"][0] not in (None, "NODEFAULT"):
            # Convert default from string to proper type.
            param["default"] = (
                python_from_type[param["type"][0]](param["default"][0]),
                param["default"][1],
            )
        elif param["value"][0] not in (None, "NODEFAULT"):
            # Use value as default, converted to proper type
            param["default"] = (
                python_from_type[param["type"][0]](param["value"][0]),
                param["default"][1],
            )
        else:
            # Use internally defined default.
            param["default"] = (
                default_from_type[param["type"][0]],
                param["default"][1],
            )

    list_attrs = [
        {
            # name.split(".")[1].lower():
            #  DET.DIT -> det__dit
            "name": name.replace(".", "__").lower(),
            #  DET.DIT -> DET DIT
            "nameeso": name.replace(".", " ").upper(),
            # TODO: Reenable description
            # 'desc': v.get("label")[0],
            "desc": v["desc"],
            "unit": v["unit"],
            "type": python_from_type[v["type"][0]].__name__,
            "default": v["default"][0],
            # TODO: Reenable defaults.
            # python_from_type[v['type'][0]](
            #     (v.get("value") if v.get("value")[0] else v.get("default"))[0]
            # )
            # TPL in name should be defined in the Template base class,
            # e.g. for TPL.START.
        }
        for name, v in params.items()
        if "TPL" not in name
    ]

    # TODO this can probably be removed
    # list_attrs.insert(
    #    0,
    #    {
    #        "name": f"dummy_{name_template_eso.lower().replace('_', '')}",
    #        "nameeso": "",
    #        "desc": "dummy",
    #        "unit": None,
    #        "type": int.__name__,
    #        "default": 42,
    #    },
    # )

    lines_attr = [
        f'    {a["name"]} = persistent("{a["desc"]} [{a["nameeso"]}][{a["unit"]}]", {a["type"]}, '
        + (f'"{a["default"]}"' if isinstance(a["default"], str) else f'{a["default"]}')
        + ")"
        for a in list_attrs
    ]
    s_attr = "\n\n".join(lines_attr)
    # Put part of the string in a separate variable so it is easier to grep
    # for generated files without finding this file.
    s_generated = "This file is automatically generated."
    s_file = f'''# -*- coding: utf-8 -*-
"""{name_template}.

{s_generated} Do not modify!
"""

from micado.main.Template import Template, persistent


class {name_template}(Template):
    """{doc_template}."""

{s_attr}'''

    print(s_file)
    return s_file


def main():
    """Generate Python classes for al TSF files."""
    fns_templates = sorted(glob.glob(f"{PATH_TEMPLATE_TSF}/*.tsf"))
    for fn_template in fns_templates:
        template_from_tsf(fn_template)


if __name__ == "__main__":
    main()
