"""Summarize CATG values.

This is a stepping stone to automatically deal with CATG values in the
CPL recipes.
"""

# For now do only standard imaging.
from micado.toolbox.create_associationmatrix import CLASSES_STANDARD_IMAGING_ALL


def print_table(dd, header):
    """Print a table."""
    barriers = ["-" * len(h) for h in header]
    separators = [tuple(["->"] * len(dd))]
    cols1 = list(zip(*dd.keys())) + separators + list(zip(*dd.values()))
    cols = [[head, bar] + list(col) for head, bar, col in zip(header, barriers, cols1)]
    sizes = [max(len(cell) for cell in col) for col in cols]
    rows = list(zip(*cols))
    for row in rows:
        scells = [f"{cell:<{size}}" for cell, size in zip(row, sizes)]
        print(" ".join(scells))


# Select only the raw and external data.
dataitems_raworext = [
    di for di in CLASSES_STANDARD_IMAGING_ALL if not hasattr(di, "recipe")
]

# Select only the processed data.
dataitems_processed = [
    di for di in CLASSES_STANDARD_IMAGING_ALL if hasattr(di, "recipe")
]

# Get all relevant dataitems.
dataitems_all = sorted(
    set(
        CLASSES_STANDARD_IMAGING_ALL
        + [di.get_base_dataitem() for di in dataitems_processed]
    ),
    key=lambda di: di.name_eso(),
)

# Each dataitem has an ESO name and a DO.CATG.
docatg_from_nameeso = {(di.name_eso(),): (di.name_oca(),) for di in dataitems_all}

# Each raw data product has a DO.CATG value that can be derived from the
# DPR keywords.
inputcatg_from_dpr = {
    (di.dpr_catg_c(), di.dpr_tech_c(), di.dpr_type_c()): (di.name_oca(),)
    for di in dataitems_raworext
}

# Each recipe has several files as input. This dictionary is a mapping
# from
# - the name of the recipe, and
# - the DO.CATG, which can be derived from the input
#   - from the SOF tag,
#   - from the PRO.CATG FITS keyword,
#   - or from the DPR FITS keywords and inputcatg_from_dpr above,
# to
# - the whether it is a RAW or CALIB input in the recipe
# - the category/tag that should be used internally in the recipe.
inputcatg_from_filecatg_and_recipe = {
    (di.recipe.prop_type.name_eso(), dep.name_oca()): (
        "RAW" if i == 0 else "CALIB",
        getattr(di.get_base_dataitem(), prop.attribute).prop_type.name_oca(),
    )
    for di in dataitems_processed
    for i, (prop, dep) in enumerate(di.get_inputdata_full().items())
}

# Each recipe has one primary output. The PRO.CATG value of this output
# depends on the DO.CATG / tag value of the primary input. This dictionary
# maps from
# - the name of the recipe,
# - the input DO.CATG (as derived above),
# to
# - the PRO.CATG value of the output.
outputcatg_from_filecatg_and_recipe = {
    (
        di.recipe.prop_type.name_eso(),
        di.get_progenitor_class().name_eso(),
    ): (di.name_oca(),)
    for di in dataitems_processed
}

if __name__ == "__main__":
    print_table(
        docatg_from_nameeso,
        header=["Name", "", "DO.CATG"],
    )
    print()

    print_table(
        inputcatg_from_dpr,
        header=["DPR.CATG", "DPR.TECH", "DPR.TYPE", "", "DO.CATG"],
    )
    print()

    print_table(
        inputcatg_from_filecatg_and_recipe,
        header=["Recipe", "CATG in", "", "Type", "Tag in"],
    )
    print()

    print_table(
        outputcatg_from_filecatg_and_recipe,
        header=["Recipe", "CATG in", "", "CATG out"],
    )

"""
Name                        DO.CATG
----                        -------
ASTROM_HDR               -> HDR_ASTROM
BACKGROUND_IMG           -> IMG_BACKGROUND
BKGFIELD_RAW             -> RAW_OPT_SKY_BKGFIELD
CALIBRATED_IMG           -> IMG_FNL_CATALOGABLE_CALIBRATED
DARK_RAW                 -> RAW_DARK
DEPERSISTEDBKGFIELD_IMG  -> IMG_DP_DPOPT_DPSKY_DPBKGFIELD
DEPERSISTEDDARK_IMG      -> IMG_DP_DPDARK
DEPERSISTEDFLAT_IMG      -> IMG_DP_DPOPT_DPFLAT
DEPERSISTEDSCIENCE_IMG   -> IMG_DP_DPOPT_DPSKY_DPSIAI_DPSCIENCE
DEPERSISTEDSTDFIELD_IMG  -> IMG_DP_DPOPT_DPSKY_DPSTDFIELD
DEPERSISTED_IMG          -> IMG_DP
DETRENDBKGFIELD_IMG      -> IMG_FNL_CATALOGABLE_DT_DTBKGFIELD
DETRENDFINALSTDFIELD_IMG -> IMG_FNL_CATALOGABLE_DTFNL_DTFNLSTDFIELD
DETRENDFINAL_IMG         -> IMG_FNL_CATALOGABLE_DTFNL_DTFNLSIAI_DTFNL
DETRENDSTDFIELD_IMG      -> IMG_FNL_CATALOGABLE_DT_DTSTDFIELD
DETREND_IMG              -> IMG_FNL_CATALOGABLE_DT
FLAT_RAW                 -> RAW_OPT_FLAT
MASTERDARK_IMG           -> IMG_MASTERDARK
MASTERFLAT_IMG           -> IMG_MASTERFLAT
PHOTOMSTDFIELD_HDR       -> HDR_PHOTOM_PHOTOMSTDFIELD
PHOTOM_HDR               -> HDR_PHOTOM
SCIENCE_RAW              -> RAW_OPT_SKY_SKYSIAI_SCIENCE
STACK_IMG                -> IMG_STACK
STDFIELD_RAW             -> RAW_OPT_SKY_STDFIELD

DPR.CATG DPR.TECH DPR.TYPE    DO.CATG
-------- -------- --------    -------
CALIB    IMAGE    DARK     -> RAW_DARK
CALIB    IMAGE    FLAT     -> RAW_OPT_FLAT
SCIENCE  IMAGE    OBJECT   -> RAW_OPT_SKY_SKYSIAI_SCIENCE
CALIB    IMAGE    PHOTOM   -> RAW_OPT_SKY_STDFIELD
CALIB    IMAGE    BCKGRND  -> RAW_OPT_SKY_BKGFIELD

Recipe                   CATG in                                    Type  Tag in
------                   -------                                    ----  ------
micado_det_depersist     RAW                                     -> RAW   RAW
micado_det_depersist     IMG_PERSISTENCE                         -> CALIB IMG_PERSISTENCE
micado_det_dark          IMG_DP_DPDARK                           -> RAW   IMG_DP_DPDARK
micado_img_flat          IMG_DP_DPOPT_DPFLAT                     -> RAW   IMG_DP_DPOPT_DPFLAT
micado_img_flat          IMG_MASTERDARK                          -> CALIB IMG_MASTERDARK
micado_img_detrend       IMG_DP_DPOPT                            -> RAW   IMG_DP_DPOPT
micado_img_detrend       IMG_MASTERDARK                          -> CALIB IMG_MASTERDARK
micado_img_detrend       IMG_MASTERFLAT                          -> CALIB IMG_MASTERFLAT
micado_img_detrend       IMG_NONLINEARITY                        -> CALIB IMG_NONLINEARITY
micado_img_background    IMG_FNL_CATALOGABLE_DT_DTBKGFIELD       -> RAW   IMG_FNL_CATALOGABLE_DT_DTBKGFIELD
micado_img_detrend_final IMG_FNL_CATALOGABLE_DT                  -> RAW   IMG_FNL_CATALOGABLE_DT
micado_img_detrend_final IMG_BACKGROUND                          -> CALIB IMG_BACKGROUND
micado_img_detrend_final HDR_STATICTABLE_ILLUMCORR               -> CALIB HDR_STATICTABLE_ILLUMCORR
micado_img_astrom        IMG_FNL_CATALOGABLE_DTFNL_DTFNLSIAI     -> RAW   IMG_FNL_CATALOGABLE_DTFNL_DTFNLSIAI
micado_img_astrom        CAT_DISTORTIONREF_ASTROMREF             -> CALIB CAT_DISTORTIONREF_ASTROMREF
micado_img_astrom        HDR_DISTORTION_DISTORTIONWAM            -> CALIB HDR_DISTORTION_DISTORTIONWAM
micado_img_astrom        HDR_DISTORTION_DISTORTIONELT            -> CALIB HDR_DISTORTION_DISTORTIONELT
micado_img_photom        IMG_FNL_CATALOGABLE_DTFNL_DTFNLSTDFIELD -> RAW   IMG_FNL_CATALOGABLE_DTFNL
micado_img_photom        HDR_ASTROM                              -> CALIB HDR_ASTROM
micado_img_photom        CAT_PHOTOMREF                           -> CALIB CAT_PHOTOMREF
micado_img_calibrate     IMG_FNL_CATALOGABLE_DTFNL_DTFNLSIAI     -> RAW   IMG_FNL_CATALOGABLE_DTFNL_DTFNLSIAI
micado_img_calibrate     HDR_PHOTOM                              -> CALIB HDR_PHOTOM
micado_img_calibrate     HDR_ASTROM                              -> CALIB HDR_ASTROM
micado_img_stack         IMG_FNL_CATALOGABLE_CALIBRATED          -> RAW   IMG_FNL_CATALOGABLE_CALIBRATED
micado_img_detrend       IMG_DP_DPOPT_DPSKY_DPSIAI_DPSCIENCE     -> RAW   IMG_DP_DPOPT
micado_img_detrend       IMG_DP_DPOPT_DPSKY_DPSTDFIELD           -> RAW   IMG_DP_DPOPT
micado_img_detrend       IMG_DP_DPOPT_DPSKY_DPBKGFIELD           -> RAW   IMG_DP_DPOPT
micado_img_detrend_final IMG_FNL_CATALOGABLE_DT_DTSIAI_DT        -> RAW   IMG_FNL_CATALOGABLE_DT
micado_img_detrend_final IMG_FNL_CATALOGABLE_DT_DTSTDFIELD       -> RAW   IMG_FNL_CATALOGABLE_DT
micado_det_depersist     RAW_DARK                                -> RAW   RAW
micado_det_depersist     RAW_OPT_FLAT                            -> RAW   RAW
micado_det_depersist     RAW_OPT_SKY_SKYSIAI_SCIENCE             -> RAW   RAW
micado_det_depersist     RAW_OPT_SKY_STDFIELD                    -> RAW   RAW
micado_det_depersist     RAW_OPT_SKY_BKGFIELD                    -> RAW   RAW

Recipe                   CATG in                     CATG out
------                   -------                     --------
micado_det_depersist     RAW                      -> IMG_DP
micado_det_dark          DEPERSISTEDDARK_IMG      -> IMG_MASTERDARK
micado_img_flat          DEPERSISTEDFLAT_IMG      -> IMG_MASTERFLAT
micado_img_detrend       DEPERSISTEDOPT_IMG       -> IMG_FNL_CATALOGABLE_DT
micado_img_background    DETRENDBKGFIELD_IMG      -> IMG_BACKGROUND
micado_img_detrend_final DETREND_IMG              -> IMG_FNL_CATALOGABLE_DTFNL_DTFNLSIAI_DTFNL
micado_img_astrom        DETRENDFINALSIAI_IMG     -> HDR_ASTROM
micado_img_photom        DETRENDFINALSTDFIELD_IMG -> HDR_PHOTOM_PHOTOMSTDFIELD
micado_img_calibrate     DETRENDFINALSIAI_IMG     -> IMG_FNL_CATALOGABLE_CALIBRATED
micado_img_stack         CALIBRATED_IMG           -> IMG_STACK
micado_img_detrend       DEPERSISTEDSCIENCE_IMG   -> IMG_FNL_CATALOGABLE_DT_DTSIAI_DT
micado_img_detrend       DEPERSISTEDSTDFIELD_IMG  -> IMG_FNL_CATALOGABLE_DT_DTSTDFIELD
micado_img_detrend       DEPERSISTEDBKGFIELD_IMG  -> IMG_FNL_CATALOGABLE_DT_DTBKGFIELD
micado_img_detrend_final DETRENDSTDFIELD_IMG      -> IMG_FNL_CATALOGABLE_DTFNL_DTFNLSTDFIELD
micado_det_depersist     DARK_RAW                 -> IMG_DP_DPDARK
micado_det_depersist     FLAT_RAW                 -> IMG_DP_DPOPT_DPFLAT
micado_det_depersist     SCIENCE_RAW              -> IMG_DP_DPOPT_DPSKY_DPSIAI_DPSCIENCE
micado_det_depersist     STDFIELD_RAW             -> IMG_DP_DPOPT_DPSKY_DPSTDFIELD
micado_det_depersist     BKGFIELD_RAW             -> IMG_DP_DPOPT_DPSKY_DPBKGFIELD
"""
