from micado.main.MDBObjectMeta import MDBObjectMeta


def get_dataitems_derived(
    dataitem: MDBObjectMeta, allowed=None, strict=False, include_self=False
):
    """Get derived dataitems.

    E.g. DetrendFinalBase -> DetrendFinal, DetrendFinalStdField,
    DetrendFinalPupil, DetrendFinalWaffle.
    """
    # Some DataItems, e.g. SKY_RAW, are abstract. List the concrete derived
    # ones instead.
    # TODO: UnspersistedSky should not allow e.g. MCD_img_cal_dark.
    #       But those figures are not included in the documentation anyway.
    list_self = [dataitem] if include_self else []
    dataitems_derived = [
        temp
        for temp in dataitem.get_derived_classes() + list_self
        if "qacits" not in temp.name_eso()
        and not temp.name_eso().endswith("base")
        and not temp.__name__.endswith("Base")
        and temp.__module__.startswith("micado.dataitems.")
    ]
    if allowed is not None:
        if strict:
            # The exact allowed classes are given
            dataitems_derived = [temp for temp in dataitems_derived if temp in allowed]
        else:
            # Cannot use 'issubclass' directly because get_derived_classes
            # might be overloaded
            dataitems_derived = [
                temp
                for temp in dataitems_derived
                if any(
                    temp in (allo.get_derived_classes() + [allo]) for allo in allowed
                )
            ]

    # TODO: figure out a better way to sort this list.
    dataitems_derived.sort(key=lambda t: t.name_eso())

    # Turns out that if there are derived dataitems, that it is then never
    # necessary to include the dataitem itself.
    # TODO: The above is not necessarily true.
    temps = dataitems_derived if dataitems_derived else [dataitem]

    # TODO: recurse on the get_derived_classes()?
    return temps


def get_progenitors_derived(dataitem: MDBObjectMeta):
    """Get full set of non-abstract derived progenitors.

    TODO: Move this to MDBObjectMeta itself.
    TODO: refactor this because it is copied from text_from_dataitem
    """
    progenitor = dataitem.get_progenitor_class()
    if progenitor is None:
        return None
    # include_self = True adds MICADO_coro_acq_star to the templates
    # of DetrendBase
    return get_dataitems_derived(progenitor, include_self=True)


def get_raws_derived(dataitem: MDBObjectMeta):
    """Get full set of derived raws."""
    dataitems_todo = [dataitem]
    dataitems_done = set()
    raws = []
    while dataitems_todo:
        dataitem_doing = dataitems_todo.pop()
        if dataitem_doing in dataitems_done:
            continue
        dataitems_done.add(dataitem_doing)
        if dataitem_doing.get_derived_classes():
            progenitors = [
                dd.get_progenitor_class() for dd in dataitem_doing.get_derived_classes()
            ]
        else:
            # get_progenitors_derived() returns a list or None
            progenitors = get_progenitors_derived(dataitem_doing) or []
        if all(prog is None for prog in progenitors):
            # Dataitem has no progenitors, it is a raw.
            raws.append(dataitem_doing)
        else:
            # Dataitem has progenitors, check those
            assert None not in progenitors
            dataitems_todo += progenitors
    return raws
